/*
  Copyright 2004, David N. Welton <davidw@dedasys.com>

  This code may be used, modified and redistributed under the same
  terms as Tcl itself.
*/

#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <tcl.h>

#define IFC_BUFLEN 8192

/*
 *-----------------------------------------------------------------------------
 *
 *  ListInterfaces --
 *
 * 	List available interfaces.
 *
 * Results:
 *	A list of interfaces.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------------
 */

static int ListInterfaces(Tcl_Interp *interp) {
    struct ifconf if_request;
    char *ifc_buffer;
    int i, j;
    int fd;
    Tcl_Obj *iflist;
    Tcl_Obj *ifname = NULL;

    /* Allocate space for the interface request. */
    ifc_buffer = (char *)malloc(IFC_BUFLEN);

    bzero(ifc_buffer, IFC_BUFLEN);
    bzero(&if_request, sizeof(if_request));
    /* We need a socket file descriptor to work with, even if it isn't
     * connected to anything. */
    fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);

    if_request.ifc_buf = ifc_buffer;
    if_request.ifc_len = IFC_BUFLEN;

    /* Fetch information from the kernel. */
    ioctl(fd, SIOCGIFCONF, &if_request);

    iflist = Tcl_NewObj();
    /* Loop through interfaces. */
    for (i = 0,  j = 0; i < if_request.ifc_len;
	 i += sizeof(struct ifreq), j++) {
	/* Fetch interface name, create a new string object. */
	ifname = Tcl_NewStringObj(if_request.ifc_req[j].ifr_name, -1);

	/* Add it to the list. */
	Tcl_ListObjAppendElement(interp, iflist, ifname);
    }
    /* Use the list as the result. */
    Tcl_SetObjResult(interp, iflist);

    /* Clean up other resources. */
    close(fd);
    free(ifc_buffer);

    return TCL_OK;
}

/*
 *-----------------------------------------------------------------------------
 *
 * GetInterface --
 *
 * 	Fetch information about a particular interface in the form of
 * 	a key/value list suitable for use as an array or dict.
 *
 * Results:
 *	A list of keys and their values.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------------
 */

static int GetInterface(Tcl_Interp *interp,
			int objc,
			Tcl_Obj *CONST objv[]) {
    int fd;

    struct ifreq ifreq;
    unsigned char *hw;
    struct sockaddr_in *sin;
    char *itf;

    char hwaddr[40];
    Tcl_Obj *addr;
    Tcl_Obj *braddr;
    Tcl_Obj *netmask;
    Tcl_Obj *result;

    if (objc != 3) {
	Tcl_WrongNumArgs(interp, 2, objv, "interface");
	return TCL_ERROR;
    }
    itf = Tcl_GetString(objv[2]);

    fd = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);

    strcpy(ifreq.ifr_name, itf);

    /* hardware address */
    ioctl(fd, SIOCGIFHWADDR, &ifreq);
    hw = ifreq.ifr_hwaddr.sa_data;
    sprintf(hwaddr, "%02x:%02x:%02x:%02x:%02x:%02x",
            *hw, *(hw + 1), *(hw + 2), *(hw + 3), *(hw + 4), *(hw + 5));

    result = Tcl_NewObj();
    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj("hwaddr", -1));
    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj(hwaddr, -1));

    /* address */
    if (ioctl(fd, SIOCGIFADDR, &ifreq) == 0) {
	sin = (struct sockaddr_in *)&ifreq.ifr_broadaddr;
	addr = Tcl_NewStringObj(inet_ntoa(sin->sin_addr), -1);

	/* broadcast */
	ioctl(fd, SIOCGIFBRDADDR, &ifreq);
	sin = (struct sockaddr_in *)&ifreq.ifr_broadaddr;
	braddr = Tcl_NewStringObj(inet_ntoa(sin->sin_addr), -1);

	/* netmask */
	ioctl(fd, SIOCGIFNETMASK, &ifreq);
	sin = (struct sockaddr_in *)&ifreq.ifr_broadaddr;
	netmask = Tcl_NewStringObj(inet_ntoa(sin->sin_addr), -1);
    } else {
	addr = Tcl_NewObj();
	braddr = Tcl_NewObj();
	netmask = Tcl_NewObj();
    }

    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj("addr", -1));
    Tcl_ListObjAppendElement(interp, result, addr);
    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj("braddr", -1));
    Tcl_ListObjAppendElement(interp, result, braddr);
    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj("netmask", -1));
    Tcl_ListObjAppendElement(interp, result, netmask);

    /* flags */
    ioctl(fd, SIOCGIFFLAGS, &ifreq);

    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj("up", -1));
    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewBooleanObj(
				 (ifreq.ifr_flags & IFF_UP)));

    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj("running", -1));
    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewBooleanObj(
				 (ifreq.ifr_flags & IFF_RUNNING)));

    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewStringObj("noarp", -1));
    Tcl_ListObjAppendElement(interp, result,
			     Tcl_NewBooleanObj(
				 (ifreq.ifr_flags & IFF_NOARP)));


    close(fd);

    Tcl_SetObjResult(interp, result);
    return TCL_OK;
}

/*
 *-----------------------------------------------------------------------------
 *
 * Ifconfig --
 *
 * 	The main function, which calls secondary functions depending
 * 	on the "subcommand".
 *
 * Results:
 *	Depends on the function requested.
 *
 * Side Effects:
 *	None.
 *
 *-----------------------------------------------------------------------------
 */

static int Ifconfig(ClientData clientData,
		    Tcl_Interp *interp,
		    int objc,
		    Tcl_Obj *CONST objv[]) {
    char *subcmd;
    if (objc < 2) {
	Tcl_WrongNumArgs(interp, 1, objv, "list | get | set");
	return TCL_ERROR;
    }

    subcmd = Tcl_GetString(objv[1]);

    if (strcmp(subcmd, "list") == 0) {
	return ListInterfaces(interp);
    } else if (strcmp(subcmd, "get") == 0) {
	return GetInterface(interp, objc, objv);
    } else {
	return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *-----------------------------------------------------------------------------
 *
 * Tclifconfig_Init --
 *
 * 	Initializes the package, creating the ifconfig command.
 *
 * Results:
 *	None.
 *
 * Side Effects:
 *	Creates ifconfig command and package.
 *
 *-----------------------------------------------------------------------------
 */

int Tclifconfig_Init(Tcl_Interp *interp) {

#ifdef USE_TCL_STUBS
    if (Tcl_InitStubs(interp, "8", 0) == NULL) {
	return TCL_ERROR;
    }
#endif

    /* Create the command. */
    Tcl_CreateObjCommand(interp, "ifconfig", Ifconfig,
			 (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

    if ( Tcl_PkgProvide(interp,
			"ifconfig",
			"0.1") != TCL_OK ) {
        return TCL_ERROR;
    }
    return TCL_OK;
}
